@@ -1,7 +1,7 @@
 use std::cmp::Ordering;
 use std::collections::VecDeque;
 
-use serde::{de, forward_to_deserialize_any};
+use serde::de;
 
 use super::seq::{ItemKind, PairSeq};
 use super::stash::{Pair, Stash};
@@ -118,7 +118,7 @@ impl<'de> PairMap<'de> {
             if let Some((key, value)) = self.parse_pair(pair)? {
                 // If it is in current level just add it as a single value
                 values.push((key, ItemKind::Value(value)));
-            } else if let Some(key) = self.stash.next_key()? {
+            } else if let Some(key) = self.stash.next_key() {
                 // Visit the stash
                 if key.is_empty() {
                     // If key is empty, then add it as a single map
@@ -176,6 +176,31 @@ impl<'de> PairMap<'de> {
 
         Ok(PairSeq::new(items))
     }
+
+    fn deserialize_invalid<V>(&self, _: V) -> Result<V::Value>
+    where
+        V: de::Visitor<'de>,
+    {
+        // We don't support raw values in the map deserializer itself, which will be triggered
+        // if query string defines a map and we expect a value
+        // The reason for not using the deserialize_any in this case is the bug in serde when using
+        // flatten
+        Err(Error::ExpectedValue)
+    }
+}
+
+macro_rules! forward_to_deserialize_invalid {
+    ($($func:ident)*) => {
+        $(
+            #[inline]
+            fn $func<V>(self, visitor: V) -> Result<V::Value>
+            where
+                V: de::Visitor<'de>,
+            {
+                self.deserialize_invalid(visitor)
+            }
+        )*
+    };
 }
 
 impl<'de> de::Deserializer<'de> for PairMap<'de> {
@@ -186,6 +211,12 @@ impl<'de> de::Deserializer<'de> for PairMap<'de> {
     where
         V: de::Visitor<'de>,
     {
+        // As we already captured all serde supported types in other methods,
+        // we only hit this if it is manually triggered either by a type's own
+        // deserialize implementation or serde's flatten and tagged enums.
+        // We may be able to guess what the underlying type is (enum, map or seq)
+        // by looking at the keys, but it results in inconsistent behaviour.
+        // May be solve in future
         visitor.visit_map(self)
     }
 
@@ -222,10 +253,79 @@ impl<'de> de::Deserializer<'de> for PairMap<'de> {
         visitor.visit_seq(&mut self.into_seq()?)
     }
 
-    forward_to_deserialize_any! {
-        bool i8 i16 i32 i64 i128 u8 u16 u32 u64 u128 f32 f64 char str string
-        bytes byte_buf unit unit_struct newtype_struct tuple
-        tuple_struct map struct identifier ignored_any
+    fn deserialize_tuple<V>(self, _: usize, visitor: V) -> Result<V::Value>
+    where
+        V: de::Visitor<'de>,
+    {
+        self.deserialize_seq(visitor)
+    }
+
+    #[inline]
+    fn deserialize_tuple_struct<V>(self, _: &'static str, _: usize, visitor: V) -> Result<V::Value>
+    where
+        V: de::Visitor<'de>,
+    {
+        self.deserialize_seq(visitor)
+    }
+
+    #[inline]
+    fn deserialize_newtype_struct<V>(self, _: &'static str, visitor: V) -> Result<V::Value>
+    where
+        V: de::Visitor<'de>,
+    {
+        visitor.visit_newtype_struct(self)
+    }
+
+    fn deserialize_unit<V>(self, visitor: V) -> Result<V::Value>
+    where
+        V: de::Visitor<'de>,
+    {
+        if self.pairs.is_empty() {
+            visitor.visit_unit()
+        } else {
+            visitor.visit_map(self)
+        }
+    }
+
+    fn deserialize_unit_struct<V>(self, _: &'static str, visitor: V) -> Result<V::Value>
+    where
+        V: de::Visitor<'de>,
+    {
+        self.deserialize_unit(visitor)
+    }
+
+    fn deserialize_map<V>(self, visitor: V) -> Result<V::Value>
+    where
+        V: de::Visitor<'de>,
+    {
+        visitor.visit_map(self)
+    }
+
+    fn deserialize_struct<V>(
+        self,
+        _: &'static str,
+        _: &'static [&'static str],
+        visitor: V,
+    ) -> Result<V::Value>
+    where
+        V: de::Visitor<'de>,
+    {
+        visitor.visit_map(self)
+    }
+
+    fn deserialize_ignored_any<V>(self, visitor: V) -> Result<V::Value>
+    where
+        V: de::Visitor<'de>,
+    {
+        visitor.visit_map(self)
+    }
+
+    forward_to_deserialize_invalid! {
+        deserialize_i8 deserialize_i16 deserialize_i32 deserialize_i64 deserialize_i128
+        deserialize_u8 deserialize_u16 deserialize_u32 deserialize_u64 deserialize_u128
+        deserialize_f32 deserialize_f64 deserialize_bool deserialize_char
+        deserialize_str  deserialize_string  deserialize_bytes  deserialize_byte_buf
+        deserialize_identifier
     }
 }
 
@@ -248,9 +348,7 @@ impl<'de> de::MapAccess<'de> for PairMap<'de> {
         }
 
         // Visit stash
-        let key = self.stash.next_key()?;
-
-        match key {
+        match self.stash.next_key() {
             Some(key) => seed
                 .deserialize(&mut Value::new(&mut Parser::new(key)))
                 .map(Some),
@@ -273,6 +371,10 @@ impl<'de> de::MapAccess<'de> for PairMap<'de> {
             }
         }
     }
+
+    fn size_hint(&self) -> Option<usize> {
+        Some(self.pairs.len())
+    }
 }
 
 impl<'de> de::EnumAccess<'de> for &mut PairMap<'de> {
@@ -298,7 +400,7 @@ impl<'de> de::EnumAccess<'de> for &mut PairMap<'de> {
             Some(key) => key,
             None => {
                 // Visit stash
-                while let Some(key) = self.stash.next_key()? {
+                while let Some(key) = self.stash.next_key() {
                     last_key = Some(key)
                 }
 
